extern "C" {
#include <stdio.h>
#include <math.h>
#include <assert.h>
}

#define NaN NAN

#include "fregr.hh"

/****************************************************************/

// we always include an autointercept

FREGR *fregr_new(int numind) {
  FREGR *gnew;
  
  gnew= (FREGR *) calloc(1,sizeof(FREGR));

  numind=numind+1;  // autointercept
  
  gnew->k=numind;
  
  gnew->d= (double *) calloc(numind+1,sizeof(double));
  gnew->thetabar= (double *) calloc(numind+1,sizeof(double));
  gnew->rbarsize= (int) ((numind+1)*(double) (numind)/2+1);
  gnew->rbar= (double *) calloc( gnew->rbarsize+1 ,sizeof(double));
  gnew->Theta= (double *) calloc(numind+1, sizeof(double));

  fregr_clear(gnew);

  return(gnew);
}


/****************************************************************
fregr_clear()
****************************************************************/
FREGR *fregr_clear(FREGR *gnew) {
  int i;

  /* gnew->k=numind; * this remains ! */
  gnew->sse= 0.0;
  gnew->n= gnew->wghtn= 0;

  for (i=0; i<=gnew->k;++i) gnew->d[i]=gnew->thetabar[i]=0.0;
  assert((gnew->k*(double) (gnew->k+1)/2) <= gnew->rbarsize);
  for (i=0; i<= gnew->rbarsize; ++i) gnew->rbar[i]=0.0;
  gnew->Theta[1]= gnew->rsq= NaN;
  
  return(gnew);
}

/****************************************************************
  fregr_end() frees all structures used in the fregression.
****************************************************************/
FREGR *fregr_end(FREGR *in) {
  if (in==NULL) return NULL; // we never had this one
  assert( in!=NULL && in->d!=NULL && in->k!=(-1) );
  free(in->d);
  free(in->thetabar);
  free(in->rbar);
  free(in->Theta);

  in->d=in->thetabar=in->rbar=NULL; in->k=(-1);
  free(in);
  return NULL;
}


/****************************************************************
void fregr_include(in, weight, yelement, xrow)
****************************************************************/
void fregr_include( FREGR *in, double weight, double yelement, const double *xrow) {
  double cbar, sbar, xi, xk, di, dprimei;
  int i, k, nextr;
  double *xcopy;
	
  /* omit observations with missing observations */
  if (isnan(yelement)) return;

  for (i=1; i<in->k; ++i) if (isnan(xrow[i])) return;  // 1 variable, means const+1 var.  so, only one should be there
  if (weight==0.0) return;

  if (weight>=0.0) ++in->n; else --in->n;  /* not checked in weighted fregression: covers -1 as deletion */

  in->wghtn+= weight;

  xcopy= (double *) malloc( sizeof(double)*(in->k+3) );
  for (i=0;i<=in->k;++i) xcopy[i+1]= xrow[i]; // autointercept
  xcopy[1]=1.0;
  xcopy[0]= xcopy[i+1]= NaN;

  for (i=1;i<=in->k;++i) {
    if (weight==0.0) { // note: weight is updated in the routine below, so this may bite!
      free(xcopy); xcopy=NULL; return; } // note: weight is updated, so leave it here!
    if (xcopy[i]<(-0.00001) || (xcopy[i]>(0.00001))) {
      xi=xcopy[i];

      di=in->d[i];
      dprimei=di+weight*(xi*xi); cbar=di/dprimei;
      sbar=weight*xi/dprimei; weight*=(cbar);
      in->d[i]=dprimei; nextr=(i-1)*((double) (2*in->k-i))/2+1;
      assert(nextr<=in->rbarsize);
      for (k=i+1;k<=in->k;++k) {
	xk=xcopy[k]; xcopy[k]=xk-xi*in->rbar[nextr];
	in->rbar[nextr]=cbar*in->rbar[nextr]+sbar*xk;
	++nextr;
      }
      xk=yelement; yelement-=xi*in->thetabar[i];
      in->thetabar[i]=cbar*in->thetabar[i]+sbar*xk;
    }
  }
  in->sse+=weight*(yelement*yelement);

  /* make sure that xcopy is known to be clobbered */
  free(xcopy); xcopy=NULL;

  /* indicate that Theta and rsq are garbage now */
  in->Theta[1]= in->rsq= NaN;
}

/****************************************************************
fregr_regress().  computes the coefficient vector
****************************************************************/
double *fregr_regress(FREGR *in) {
  register int i, k, nextr;

  if (in->n < in->k) {
    for (i=0; i<=in->k; ++i) in->Theta[i]= NaN;
  }

  for (i=(in->k); i>=1; --i) {
    in->Theta[i]= in->thetabar[i]; nextr=(i-1)*((double) (2*in->k-i))/2+1;
    assert(nextr<=in->rbarsize);
    for (k=i+1;k<=in->k;++k) {
      in->Theta[i]-=(in->rbar[nextr]*in->Theta[k]);
      ++nextr;
    }
  }
  
  return in->Theta;
}


/****************************************************************/


#ifdef TEST

int main() {
  FREGR *g= fregr_new( 1 );

  for (int i=1; i<=10; ++i) {
    double x[]= { NaN, i*i, NaN };
    fregr_include(g, log( 10.0+(double)i ), i, x);
  }
  g->print();

  double *beta= fregr_regress(g);
  printf("%f %f\n", beta[1], beta[2]);
  assert( fabs( beta[2] - 0.08522 ) < 0.0001 );

  fprintf(stderr, "----------------------------------------------------------------\n");
  fregr_clear(g);

  double scaler=1.0;

  for (int i=1; i<=10; ++i) {
    double x[]= { NaN, i*i, NaN };
    fregr_include(g, scaler*log( 10.0+(double)i )/10.0, i, x);

    if (i==5) { scaler= 10.0; g->rescale( 10.0 ); }
  }
  g->print();

  beta= fregr_regress(g);

  printf("%f %f\n", beta[1], beta[2]);
  assert( fabs( beta[2] - 0.08522 ) < 0.0001 );



  return 0;
}

#endif
